#ifndef FER220_H
#define FER220_H

/*  Fermat 64 bit computer algebra system for Linux.
    Copyright (C) 2020 Robert H. Lewis.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.  */


/* May 2003.  Attempt to split old fer220 by putting all poly and quoly arithmetic here.  We wouldn't
   neeed the quoly except we still want an interrupt facility.  That means we need a procedure
   eerror here, along with its own echeck_cmd_keys.  'e' means easy.  We also need eparse, e_old_expre,
   etc.  efact is much restricted. */
/* new version 3.3 Jan 2004 */

#include "ferm120.h"
#include "fer320.h"

#include <stdint.h> 

#define ufgets(a,b,c) if (!fgets((char *)a,b,c)) fermat_error("fgets returns null",false) /*AAR*/
#define ustrncmp(a,b,c) strncmp((char *)a,(char *)b,c) /*AAR*/
#define ustrcat(a,b) strcat((char *)a,(char *)b)  /*AAR*/

typedef int32_t ymills[5];
  
extern void find_total_available_memory(void); /*AAR */
/* extern uint32_t memavail(void);   Oct2 2010 moved 
extern uint64_t totalavailablememory;   */

extern int64_t ddumm1;  /* Aug 2013 */
extern uint32_t ddumm2;
extern uint32_t ddumm3;


extern int32_t c_cntdepth;
extern int32_t c_cntcount;
extern int32_t a_cntdepth;
extern int32_t a_cntcount;
extern int32_t g_cntcount;
extern int32_t d_cntdepth;
extern int32_t d_cntcount;
/* extern int32_t hgcd_cut; not here!  Jul 2010 */

extern Boolean linesuppression;  /* Jun 2005 */

/* Dec 2001  counts and depths for all the content procs. Nov 2002 */
/* extern int32_t emgcd_basic_cut;     Oct 2003  This is defined in fer320.h!  Remove it here.  Jan 2021 */
extern int16_t promptsize;
extern int16_t tripval;
extern nameptr startexpr;
extern nameptr prev_expr;
extern nameptr currentchar;
extern nameptr zilch;
extern condexptr last_good_one;
extern nameptr last_attached;
extern comptr last_comm;
extern shortstring leftover;
extern shortstring saveleft;
// extern tokentipe blank;  moved to fer320  Nov 2018
// extern tokentipe semicolon;  moved to 320.h May 2022
extern tokentipe zero;
extern tokentipe multsign;
extern tokentipe plussign;
extern tokentipe quotsign;
extern tokentipe sqrtsign;
extern tokentipe expsign;
extern tokentipe factsign;
extern tokentipe diag_sign;
extern tokentipe comma;
extern tokentipe minussign;
extern tokentipe esign;
extern tokentipe concatsign;
extern tokentipe logsign;
extern tokentipe greatintsign;
extern tokentipe quotesign;
extern tokentipe polysign;
extern tokentipe modoutsign;
extern tokentipe current_func;
extern tokentipe empty;
extern tokentipe bulletsign;
extern tokentipe arr_arrsign;
extern tokentipe sys_func_tok;
extern tokentipe sys_func_token;

extern formptr sys_func;

extern test_array base_7_list;
extern test_array base_3_list;
extern test_array base_2_list;

extern Boolean monomal[max_number_polys + 1];     /* for Hamilton path */

extern int32_t iii;   // Sep 2015.  keep for debugging.

extern void warning(int32_t x, nametype bad); /*AAR was nametype *bad*/

extern void resume_old_polynames(void);

extern Boolean need_to_turn_off_mod_out(Boolean allpolys, Boolean want_LaGrange, 
                                Boolean do_gb, int32_t mycut, int32_t m);  /* Jan 2004 */

extern void adjust_polyname_ptrs(void);  /* Jan 2004 */

extern void echeck_cmd_keys(void);

extern void listchars(textptr outf, nameptr p, Longiptr ind, Boolean *afterspace, Boolean itsacond);

extern void listrht(textptr outf, Boolean outf_is_tty, condexptr p, comptr comm, int32_t ind);

extern void listform(textptr outf, Boolean outf_is_tty, formptr p, int32_t ind);

extern void save_input_stream(nameptr *p);

extern void out_err_mess(int32_t x);    // Sep2 2016, kill eerror

/*was bigstring *messge -- AAR*/
extern void dumpitt(/*bigstring*/char * messge, int32_t n, Boolean newline, Boolean multin, Boolean colon);

/*was bigstring *mess -- AAR*/
extern void fermat_error(/*bigstring*/char * mess, Boolean push);  /* Mar 2009 */

extern void reset_pseudo_poly(void);

extern void do_prompts(void);

extern void set_quotient(quoly m, textptr inf, nameptr *flag, int32_t errno, char end_char);

extern void scanner(textptr inf, tokentipe *tok, nameptr *flag);

extern void better_sparse_display(textptr outf, arrptr p);  /* Dec 2007 */

extern void sparse_display(textptr outf, arrptr p, Boolean istty);

extern Boolean all_small(save_EnvRec *save_Env);  /* moved Jul 2007 */

extern void do_array_p(save_EnvRec *save_Env);  /* Jul 2007 */

extern Boolean do_sorted_list(save_EnvRec *save_Env);  /* Jul 2007 */

extern void save(textptr inf, nameptr *flag, bigstring newname); /* Jul 2007 */

extern void arrdis(textptr outf, tokentipe *x, int32_t location, int32_t rows, int32_t cols, int32_t total_rows, int32_t desired); /* Jul 2007 */

extern void arrdisplay(textptr outf, Boolean outf_is_tty, Boolean pretty, tokentipe *x, arrptr q, int32_t 
    location, int32_t rows, int32_t cols, int32_t firstrow, int32_t firstcol, int32_t total_rows, int32_t desired, Boolean 
    charprint, Boolean from_store, Boolean extras, workptr stash);

extern void remove_args(funcptr f, Boolean cant_call_error);

/* extern void poly_display(textptr outf, Boolean outf_is_tty, Boolean flag, poly p, Longiptr mylength,
    tokentipe *endchar, polypt q, Boolean extra_space, Boolean maybe_just_number, Longiptr lines);
   moved 320.h May 2022 */

extern void adjlength(Longiptr mylength, int32_t n);

extern void quoly_display(textptr outf, Boolean outf_is_tty, Boolean flag, Boolean linend, quoly p, 
    Longiptr mylength, tokentipe *endchar, polypt q, Boolean extra_space, Boolean leadoff, Boolean 
    char_flag);

extern void insert_(tokentipe x,quolynode *num,varptr q);

extern void new_readin(textptr inf, imptr n, nameptr *flag, Longiptr outcount);

extern void get_decimal_part(quolyptr n, textptr inf, nameptr *flag);

extern void eget_exponent(textptr inf, nameptr *flag, imptr m2, Boolean factor);

extern void efact(textptr inf, quolyptr *qp, quolyptr n, colm_ptrptr *aux, Longiptr colpos, 
    nameptr *flag, Boolean leave_it);

extern void e_old_fact(textptr inf, quolyptr n, nameptr *flag);

extern void e_old_term(textptr inf, quolyptr n, nameptr *flag);

extern void e_old_expre(textptr inf, quolyptr n, nameptr *flag);

extern void eparse(textptr inf, quolyptr n, nameptr *flag);

extern void sub(imptr n1, imptr n2, imptr n3);

extern void fold_poly(polyptrr p, polyptrr modder, int32_t expon, polypt pp);

extern void actual_eval_poly(polyptrr n, polyptrr nden, polyptrr n2, polyptrr d2, int32_t target,
                             int32_t source, polypt pt);

extern void change_poly_modout_data(void);

extern void poly_change_to_mod(polyptrr p, Boolean shouldwarn);

extern void number_power(imptr n, int32_t expont);

extern void total_poly_modout(polyptrr p, polypt pt);

extern void lint_quoly_compute_power(quolyptr q, int32_t expont);

extern void compute_power(polyptrr n, int32_t expont, polypt pol);

extern void quoly_compute_power(quolyptr q, int32_t expont);

extern int32_t check_and_convert(imptr n2);

extern void raise_to_power(quolyptr n, imptr n2);

extern int32_t imod_inverse(int32_t seca);

extern void mod_inverse(int32_t seca, Longiptr ans);

extern void quot(imptr n1, imptr n2, imptr n3);

extern void gen_poly_divide(poly n, poly m, polyptrr q, polypt pt, Boolean *failed);  /* Jan 2006 */

extern void fast_poly_divide(polyptrr n, poly m, polypt pt, int64_t xx, Boolean *worked);   /* Jan 2006  Jul 2018 */

extern void fast_mod_and_div(poly n, poly m, polyptrr r, polyptrr qt, polypt pt, Boolean from_basic);

extern void mod_and_div(imptr n, imptr n2, imptr n3, imptr n4);

extern void num_pseudo_div(poly n, poly m, polyptrr q, polyptrr r, polyptrr coef, polypt pt);

extern void poly_mod_invert(poly q, quolyptr p, polypt pp);

extern Boolean trial_poly_divide(poly n, poly m, polyptrr q, polypt pt, Boolean shortcut,
                                 Boolean yes_check, Boolean kill_n); 

extern void fast_divide(polyptrr p, poly d, polypt pt, int32_t xx);  /* Jan 2006 */

extern void mod_or_div(imptr n, imptr n2, imptr n3, char which, Boolean my_posmod);

extern void poly_mod_or_div(poly n, poly m, polyptrr p, char which, polypt pt);

extern void quoly_mod_or_div(quoly n, quoly m, quolyptr p, char which, Boolean p_equals_n);   // Mar 2022

extern void mod_trial_poly_divide(poly x, poly y, Boolean fast_yes, Boolean *worked, Boolean
                                  *checked, polypt pt);

extern void content(poly q, polyptrr r, polypt pt, Boolean verbon,   /* Oct 2002 */
                    Boolean allcons, Boolean qsave, int32_t code);  // Apr 2020

extern void monic_poly(polyptrr q, polyptrr r, polypt pt);

extern void poly_mod_prime(polyptrr n, int32_t expont, polypt pol);

extern void basic_poly_gcd(poly n, poly m, polyptrr p, Boolean need_monic, polypt pt);

extern poly Chinese(poly val, poly modder, numptr q, polypt pt);

extern void pth_root(polyptrr p, polypt pt);

/*extern void general_content(poly p, int16_t depth, polyptrr r, polypt pt, Boolean *wantr, Boolean wantcons); Mar 2006 moved May 2022 to 320.h*/

extern void polygcd(poly u, poly v, polyptrr r, Boolean leave_early, Boolean do_fast, Boolean
             allcons, polypt pt, Boolean usave, Boolean vsave, Boolean canreset, Boolean *found_something,
             Boolean callzip); /* Feb 2011, Sep 2016 */

extern void Lpolygcd(polyptrr u, polyptrr v, polyptrr r, Boolean v_odd, polypt pt);

extern void change_num(imptr n, Boolean should_free, Boolean shouldwarn);

extern void quoly_mod_convert(quoly q);

extern int32_t inverse_prod(numptr h);

extern void poly_diff(polyptrr n, int32_t degree);

extern void general_poly_diff(polyptrr p, int32_t degree, int16_t height);

extern void square_free_factor(poly x, polyptrr sq, Longiptr count, polypt pt);

extern void factorial(imptr n);

extern void poly_mod_and_div(poly n, poly m, polyptrr r, polyptrr qt, polypt pt);

extern void mod_out_quoly(quolyptr q, polyptrr modder, polyptrr alter, polyptrr md, int32_t depth, 
    int32_t expon, polypt qq);

extern void above_field_modout(polyptrr p);

extern Boolean iszero(imptr num);

extern Boolean trial_poly_divide_noq(poly n, poly m, polypt pt, Boolean shortcut, Boolean yes_check, Boolean kill_n);

// Feb 2023 moved to fer320.h extern void better_poly_add(poly n, poly m, polyptrr p);

extern void best_poly_add(poly n, poly m, polyptrr p);

extern void scalar_by_poly(imptr x, poly y, polyptrr q);

// Jan 2023 move to fer320.h extern void poly_add(poly n, poly m, polyptrr p);

extern void poly_sub(poly n, poly m, polyptrr p);

extern void better_poly_sub(poly n, poly m, polyptrr p);

extern void best_poly_sub(poly n, poly m, polyptrr p);

extern void poly_mult(poly n, poly m, polyptrr p, polypt pol);

extern void convert_to_mono(poly p, monoptr qh, monoptr qt, polypt pt, __uint128_t  prevexp, int32_t *size);  /* Jul 2013 */

extern void poly_quot(poly n, poly m, polyptrr p, int32_t xx);  /* Mar 2010 */

extern void basic_poly_divide(poly n, poly m, polyptrr q, polyptrr r, Boolean cancall);

extern void special_div(polyptrr u, polyptrr s, polypt pt);

extern void gcd_coefs(poly q, polyptrr r);

extern void quoly_reciprocal(quolyptr n);

extern void reduce(polyptrr n, polyptrr m, polyptrr com, polyptrr num, polyptrr den, Boolean el_n, 
    Boolean el_m, Boolean nflag, Boolean dflag);

extern void check_quoly(quolyptr q, Boolean q_odd);

extern void lint_normal_form(quoly q, Boolean laur);

extern void sanitize(quoly q);

extern void normal_form(quoly q, Boolean laur);

extern void listvars(Boolean wantvalues);           /* August 1996 */

extern void dump(textptr outf, Boolean outf_is_tty, Boolean flag, Boolean mynoisy);   /* August 1996 */

extern void listfuncs(textptr outf, textptr inf, Boolean outf_is_tty, tokentipe *x, nameptr *flag);

extern void add_numb_rat(quoly n, quoly m, quolyptr p, Boolean add);

extern void add_rats(quoly n, quoly m, quolyptr p, Boolean add);

extern void mult_numb_rat(quoly n, quoly m, quolyptr p);

extern void mult_rats(quoly n, quoly m, quolyptr p);

extern void quot_rat_numb(quoly n, quoly m, quolyptr p);

extern void quot_rats(quoly n, quoly m, quolyptr p);

extern void best_quoly_add(quolyptr n, quoly m);

extern void fast_quoly_mult(quolyptr n, quolyptr m);

extern void best_quoly_sub(quolyptr n, quoly m);

extern void fast_gen_quoly_add(quolyptr n, quolyptr m, Boolean add);

extern void fast_gen_quoly_mult(quolyptr n, quolyptr m, Boolean laur);

extern void fast_quoly_quot(quolyptr n, quolyptr m, Boolean nogcd);

extern void no_denom(poly r, quoly s, quolyptr q, int16_t flag);

extern void no_denom_minus(poly r, quoly s, quolyptr q);  /* Oct 2007 */

extern void no_denom_minus_2(poly r, quoly s, quolyptr q);  /* Oct 2007 */

extern void crude_poly_disp(textptr outf, Boolean outf_is_tty, poly p, int32_t depth, polypt q);  /* Nov 2005 */

extern void spec_quoly_mult(quoly pv, quoly dummy, quolyptr prod, polypt pp2);
  /*added -- AAR*/

extern Boolean mon_less(poly p, poly q, Boolean *equal); /* Jul 2006 */

extern void sparse_mod_redrowech(Boolean reduce, quolyptr n, sparseptr *spar, boolarray *want, int32_t rows, int32_t cols,
            spar_arr *auxspar, int32_t stop, Boolean set_to_5, int32_t source);  // Apr2 2016  Sep 2016  move here  Oct 2016 Dec 2016 Jul 2017 Feb 2018

extern Boolean quoly_mon_less(quoly p, quoly q); /* Sep 2010 */

extern Boolean quoly_mon_less_eq(quoly p, quoly q); /* Sep 2010 */

extern void splice_poly(quolyptr p, quolyptr q, int32_t m);  /* Aug 2010 */   /* Aug 2011 did not have extern !?!?! */

/* extern void mod_number_power(int32_t n, int32_t e, Longiptr ans);   Aug 2011   moved out Sep2 2016 */

extern void field_reduce(polyptrr n, polyptrr m, polyptrr com, polyptrr num, polyptrr den, polyptrr mu, Boolean el_n, Boolean el_m, Boolean nflag, Boolean dflag);  // Jun 2016

extern void general_quoly_add(quoly n, quoly m, quolyptr q, Boolean add);  // Sep 2016 moved here

extern void gen_quoly_mult(quoly n, quoly m, quolyptr q, Boolean laur);   // Sep 2016 moved here

extern void quoly_mult(quoly n, quoly m, quolyptr p);   // Sep 2016 moved here

extern void zcol(quolyptr qp, arrptr q, lintptr num, int32_t cols, boolarray *want, workptr arr);  /* Mar3 2016 Sep 2016 */

extern void zip_zcol(sparseptr *sq, int32_t *num, int32_t *rows, int32_t cols, sparseptr *qz);  /* Sep 2016 */
													
extern void better_quoly_add(quoly n, quoly m, quolyptr p);   // Sep 2016  move here

extern void better_quoly_sub(quoly n, quoly m, quolyptr p);   // Oct 2016  move here

extern void extract_coef(poly n, polyptrr s, int16_t height, int32_t expo, Boolean zeroexp);  // Oct 2016

extern polyptrr poly_z_num_coef(polyptrr pp);   // Oct 2016

extern void alt_mod_out_quoly(quolyptr q, polyptrr modder, polyptrr alter, polyptrr md, int32_t depth, int32_t expon,
                   polypt qq);  // Nov 2016
									 
extern void egcd_polys(poly n, poly m, polyptrr gd, textptr inf, nameptr *flag, Boolean flip);  // Nov 2016

extern void egcd_nums(quolyptr n, quolyptr m, textptr inf, nameptr *flag);   // Nov 2016

void egcd_num_poly(quolyptr n, quolyptr m, textptr inf, nameptr *flag, Boolean flip);  // Nov 2016

extern void polylist(textptr outt, Boolean saveflag);   // Nov 2016

extern void get_from_array(textptr inf, nameptr *flag, bigstring newname);  // Nov 2016

extern void dumbscanner(textptr inf, tokentipe *tok, nameptr *flag, Boolean *endline);  // Nov 2016

extern int32_t bigech(quolyptr n, sparseptr *eeq, int32_t req, int32_t cols, Boolean set_to_5, poly *qlist );  // Nov 2016

extern void spec_check_set_first_col(sparseptr sp, int32_t *j, int32_t r);  // Nov 2016

extern void quick_sparse_display(textptr outf, sparseptr sp, int32_t rows, int32_t base);  // Nov 2016

extern void free_arr(workptr ar, arr_descr *arf);   // Jan 2017

extern void free_work(void);  // Jan 2017
                   
extern int32_t terms_in_coef(poly n, int16_t height, int32_t expo); // Feb 2017

extern void simplest_poly_add(poly n, poly m, polyptrr p);  // Feb 2018

extern void spec_general_quoly_add(quoly n, quoly m, quolyptr q, Boolean add);  // Feb 2018

extern void enew_worknode(Longiptr p, tokentipe *x);  // Aug 2018

extern void mat_invert(workptr ar, workptr outar, Boolean ar_free, Boolean oar_free, int32_t m, quolyptr n, Boolean invert, Boolean equal,
    sparseptr *spar, poly fct, int32_t the_degree, int32_t the_degree2, int32_t the_degree3, int32_t the_power, Boolean really, int32_t numprimes);
    /* Jan 2008, Oct 2009  moved here Aug 2018  Jan 2021 */

extern void insert_var(tokentipe *x, quolynode *num, varptr *q);  // Aug 2018

extern void call_total_poly_modout(polyptrr p, polypt pt);  // Nov 2018

extern void simple_monic_poly(poly q);  // Jan 2020

// extern void B(int32_t m);  // Jan 2021  Mar 2022 move to 320

extern void mod_poly_at_numb(polyptrr p, poly q, poly n2, polypt pol, Boolean addpad);  // Feb 2021

// move to fer320.h Mar 2022 extern void case_one(polyptrr p, polyptrr pden, int16_t target, poly n2, poly d2, polypt pol);

extern void highest_level(polyptrr q, polyptrr qden, poly n2, poly d2, int16_t level, polypt pol);   // Nov 2021

extern void f_egcd_polys(poly n, poly m, polyptrr gd, textptr inf, nameptr *flag, Boolean flip);  // Nov 2021

extern void simplest_poly_sub(poly n, poly m, polyptrr p);  // Dec 2021

extern void most_of_f_egcd(poly n, poly m, polyptrr gd, Boolean flip, polyptrr xx1, polyptrr yy1, Boolean trip_error);  // Dec2 2021 Apr 2023

extern void more_general_content(poly p, int16_t depth, polyptrr r, polypt pt, Boolean *wantr, Boolean wantcons);  // Mar 2022

#endif
